import Prelude hiding (lookup)

data BBTree a = Nil          -- empty tree
              | Node         -- tree node
                  !(BBTree a)  -- left subtree
   {-# UNPACK #-} !Int         -- size of this tree
                  a            -- element stored in the node
                  !(BBTree a)  -- right subtree
                deriving (Eq, Ord, Read, Show)

{-# INLINE size #-}
size Nil = 0
size (Node _ s _ _) = s

left (Node l _ _ _) = l
right (Node _ _ _ r) = r
key (Node _ _ k _) = k
eval Nil = ()
eval (Node l _ x r) = x `seq` eval l `seq` eval r

{-# INLINE node #-}
node :: BBTree a -> a -> BBTree a -> BBTree a
node left key right = Node left (size left +1+ size right) key right

{-# INLINE lookup #-}
lookup :: Ord a => a -> BBTree a -> Maybe a
lookup k Nil = Nothing
lookup k (Node left _ key right) =
  case k `compare` key of
    LT -> lookup k left
    EQ -> Just key
    GT -> lookup k right

{-# INLINE insert #-}
insert :: Ord a => a -> BBTree a -> BBTree a
insert k Nil = node Nil k Nil
insert k (Node left _ key right) =
  case k `compare` key of
    LT -> balanceL (insert k left) key right
    EQ -> node left k right
    GT -> balanceR left key (insert k right)

{-# INLINE delete #-}
delete :: Ord a => a -> BBTree a -> BBTree a
delete _ Nil = Nil
delete k (Node left _ key right) =
  case k `compare` key of
    LT -> balanceR (delete k left) key right
    EQ -> glue left right
    GT -> balanceL left key (delete k right)

{-# NOINLINE glue #-}
glue Nil right = right
glue left Nil = left
glue left right
  | size left > size right =
    let (key', left') = extractMax left
    in node left' key' right
  | otherwise =
    let (key', right') = extractMin right
    in node left key' right'

{-# NOINLINE extractMin #-}
extractMin (Node Nil _ key right) = (key, right)
extractMin (Node left _ key right) = case extractMin left of
  (min, left') -> (min, balanceR left' key right)

{-# NOINLINE extractMax #-}
extractMax (Node left _ key Nil) = (key, left)
extractMax (Node left _ key right) = case extractMax right of
  (max, right') -> (max, balanceL left key right')

{-# INLINE fold #-}
fold :: (b -> a -> b) -> b -> BBTree a -> b
fold f z tree = go z tree
  where go z Nil = z
        go z (Node left _ key right) = go (f (go z left) key) right

{-# NOINLINE join #-}
join :: BBTree a -> a -> BBTree a -> BBTree a
join Nil key right = insertMin key right
  where insertMin key Nil            = Node Nil 1 key Nil
        insertMin key (Node l _ k r) = balanceL (insertMin key l) k r

join left key Nil = insertMax key left
  where insertMax key Nil            = Node Nil 1 key Nil
        insertMax key (Node l _ k r) = balanceR l k (insertMax key r)

join left@(Node ll ls lk lr) key right@(Node rl rs rk rr)
  | ls `much_bigger_than` rs = balanceR ll lk (join lr key right)
  | rs `much_bigger_than` ls = balanceL (join left key rl) rk rr
  | otherwise                = node left key right

{-# INLINE split #-}
split :: Ord a => a -> BBTree a -> (BBTree a, BBTree a)
split _ Nil = (Nil, Nil)
split key (Node l _ k r) =
  case compare key k of
    LT -> case split key l of (l', r') -> (l', join r' k r)
    EQ -> (l, r)
    GT -> case split key r of (l', r') -> (join l key l', r)
